/*
  Rev 2 of Wifi_IR_Remote_TeensyLC

  Load this sketch to the TeensyLC module 

  Rev2 removes IR_Receive to separate sketches IRrecorder_AnalysIR_TeensyLC.ino and IRrecorder_Raw_TeensyLC.ino
  Load one of those sketches to capture your existing remotes IR signal.

*/

/**
   Wifi_IR_Remote_TeensyLC
   https://www.forward.com.au/pfod/HomeAutomation/Wifi_IR_Remote/index.html

   (c)2017 Forward Computing and Control Pty. Ltd.
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.
*/

#include <stdint.h>
#include "cmdDefines.h"
#include "BeyondWizIRcodesDefines.h"
#include "PanasonicTVIRcodes.h"

const uint16_t DEFAULT_REPEAT_INTERVAL = 250; // in mS, used when repeatInterval in IR_DATA struc is 0

// add extra enums here when you add more of the protocols supported by IRremote
enum codeTypesEnum {RAW_code, LIR_code, NEC_38_code, NEC_38_REV8_code, PANASONIC38_4004_code, PANASONIC38_4004_POWER_MACRO_code};
// note PANASONIC38_4004_POWER_MACRO_code is hardcoded to send TV_POWER + TV_MUTE


// IR_DATA structure store the data for each code, macros (below) are used to populate the arrays of IR_DATA structures
struct IR_DATA {
  char *codeName;
  uint8_t* data;
  unsigned int len;  // array length in bytes always use 0 if data is not an array
  uint16_t repeatInterval; // in mS, if this is zero then this code does NOT repeat
  enum codeTypesEnum codeType;  // a codeTypesEnum
  char cmd;       // the cmd that triggers this code send one char only -- usually a to z and A to Z only
};

unsigned long repeatTimer = 0;
bool repeatTimerRunning = false; // set true when repeating

unsigned int mode = 0; // default is IR_Remote_TX
// mode 1 == USB commands   for testing Teensy IR output.  Type in the commands that would normally come from the WiFi (via hardware Serial1) into the IDE monitor
// mode 2 == pass through   for programming ESP8266 and checking the IR commands being sent from the WiFi to Teensy

Stream* cmdInput = &Serial1; // set to either Serial or Serial1, default Serial1
/**********************************************
  // ======== examples of codes
  // NEC_REV codes
  uint32_t stb_mute = 0xBE000CF3;

  // RAW
  // NOTE first value is the frequency so need to edit the output generated from PoorMans_FCC and from AnalysIR IRremote raw batch output
  // and add the frequency to front of the array  i.e. for Modulation Frequency: 38324 (Hz)  == 38khz
  // 8976, ..
  // becomes
  // 38, 8976  as shown below
  unsigned int stb_raw_mute[] = {38, 8976, 4520, 552, 580, 528, 1708, 528, 1708, 556, 1704, 532, 1708, 552, 1688, 548, 580, 524, 1712, 552, 576, 536, 572, 552, 576, 528, 576, 556, 580, 524, 580, 548, 560, 548, 580, 524, 580, 552, 576, 532, 1708, 552, 1708, 528, 576, 528, 584, 548, 576, 532, 576, 552, 1708, 528, 1708, 552, 556, 552, 580, 524, 1708, 556, 1704, 532, 1708, 552, 1688, 548};

  // AnalysIR LIR code
  char stb_lir_mute[] =  "LIR: 26 2308 11AC 0220 0244 0212 06AE 01 F2 51 2F 1A 22 F1 42 21 1F 24 1F";  // NOTE for LIR code must be a null terminated string
  // NOTE LIR supports pulse widths >65535 BUT these will only work if running on a cpu where uint32_t <= unsigned int
  // on Uno and Mega2650 unsigned int == uint16_t so very long pulses will not work
  // TeensyLC unsigned int == uint32_t so very long pulses will work.

  // first match of cmd wins if there are duplicates
  struct IR_DATA codes[] = { // note need to have the definitions in same order as struct
  {.codeName = "stb_mute", .data = (uint8_t*)&stb_mute, .len = 0, .codeType = NEC_38_REV8_code, .cmd = 'a'}
  , {.codeName = "stb_raw_mute", .data = (uint8_t*)&stb_raw_mute, .len = sizeof(stb_raw_mute), .codeType = RAW_code, .cmd = 'a'}
  , {.codeName = "stb_lir_mute", .data = (uint8_t*)&stb_lir_mute, .len = sizeof(stb_lir_mute), .codeType = LIR_code, .cmd = 'b'}
  };
*************************************/

//  codes for  Menu dwg 'A'  stored in IR_DATA_A array
//  codes for  Menu dwg 'B'  stored in IR_DATA_B array


// note need to have the definitions in same order as struct
#define lirEntry(X, Y)  {.codeName = #X, .data = (uint8_t*)&X, .len = sizeof(X), .repeatInterval = 0, .codeType = LIR_code, .cmd = Y}

#define necRevEntry(X, Y)  {.codeName = #X, .data = (uint8_t*)&X, .len = 0, .repeatInterval = 0, .codeType = NEC_38_REV8_code, .cmd = Y}
#define necRevRepeatEntry(X, Y, R)  {.codeName = #X, .data = (uint8_t*)&X, .len = 0, .repeatInterval = R, .codeType = NEC_38_REV8_code, .cmd = Y}

//#define tvEntry(X, Y)  {.codeName = #X, .data = (uint8_t*)&X, .len = sizeof(X),  .repeatInterval = 0, .codeType = LIR_code, .cmd = Y}
//#define tvRepeatEntry(X, Y, R)  {.codeName = #X, .data = (uint8_t*)&X, .len = sizeof(X), .repeatInterval = R, .codeType = LIR_code, .cmd = Y}

#define tvEntry(X, Y)  {.codeName = #X, .data = (uint8_t*)&X, .len = 0,  .repeatInterval = 0, .codeType = PANASONIC38_4004_code, .cmd = Y}
#define tvRepeatEntry(X, Y, R)  {.codeName = #X, .data = (uint8_t*)&X, .len = 0, .repeatInterval = R, .codeType = PANASONIC38_4004_code, .cmd = Y}
#define tvPowerMacro(X,Y)  {.codeName = #X, .data = (uint8_t*)&X, .len = 0,  .repeatInterval = 0, .codeType = PANASONIC38_4004_POWER_MACRO_code, .cmd = Y}
// PANASONIC38_4004_POWER_MACRO_code sends TV_MUTE as well

// first match of cmd wins if there are duplicates
struct IR_DATA codes_A[] = { // note need to have the definitions in same order as struct
  necRevEntry(PVR_OK, pvr_okCmd_A)  //{.data = (uint8_t*)&PVR_OK, .len = 0, .codeType = NEC_38_REV8_code, .cmd = pvr_okCmd_A}
  , necRevRepeatEntry(PVR_LEFT, pvr_leftCmd_A, DEFAULT_REPEAT_INTERVAL)
  , necRevRepeatEntry(PVR_RIGHT, pvr_rightCmd_A, DEFAULT_REPEAT_INTERVAL)
  , necRevRepeatEntry(PVR_UP, pvr_upCmd_A, 550)
  , necRevRepeatEntry(PVR_DOWN, pvr_downCmd_A, 550)
  , tvPowerMacro(TV_POWER,tv_powerCmd_A) //  tvPowerMacro sends TV_POWER + TV_MUTE  regardless of the value of the first parameter here
  , tvEntry(TV_MUTE, tv_muteCmd_A)
  , necRevEntry(PVR_FILE_PLAY, pvr_recordingsCmd_A)
  , necRevEntry(PVR_EXIT, pvr_exitCmd_A)
  , necRevEntry(PVR_PLAY, pvr_playCmd_A)
  , necRevEntry(PVR_STOP, pvr_stopCmd_A)
  , necRevEntry(PVR_PAUSE, pvr_pauseCmd_A)
  , necRevEntry(PVR_FF, pvr_fastForwardCmd_A)
  , necRevEntry(PVR_REW, pvr_fastReverseCmd_A)
  , necRevEntry(PVR_RED, pvr_redCmd_A)
  , necRevEntry(PVR_GREEN, pvr_greenCmd_A)
  , necRevEntry(PVR_YELLOW, pvr_yellowCmd_A)
  , necRevEntry(PVR_BLUE, pvr_blueCmd_A)
  , tvEntry(TV_AV, tv_avCmd_A)
  , necRevRepeatEntry(PVR_VOL_UP, pvr_VolUp_A, DEFAULT_REPEAT_INTERVAL)
  , necRevRepeatEntry(PVR_VOL_DOWN, pvr_VolDw_A, DEFAULT_REPEAT_INTERVAL)
  , tvRepeatEntry(TV_VOL_UP, tv_VolUp_A, DEFAULT_REPEAT_INTERVAL)
  , tvRepeatEntry(TV_VOL_DOWN, tv_VolDw_A, DEFAULT_REPEAT_INTERVAL)
  , necRevEntry(PVR_GUIDE, pvr_guideCmd_A)
  , necRevEntry(PVR_POP_UP, pvr_popUpCmd_A)
  , necRevEntry(PVR_REC, pvr_recCmd_A)
  , necRevEntry(PVR_SUBTITLE, pvr_subTitlesCmd_A)
};

// first match of cmd wins if there are duplicates
struct IR_DATA codes_B[] = { // note need to have the definitions in same order as struct
  necRevEntry(PVR_NEXT, pvr_jumpForwardCmd_B) // // {.data = (uint8_t*)&PVR_NEXT, .len = 0, .codeType = NEC_38_REV8_code, .cmd = pvr_jumpForwardCmd_B}
  , necRevEntry(PVR_PREV, pvr_jumpReverseCmd_B)
  , necRevEntry(PVR_SLOW_FWD, pvr_slowForwardCmd_B)
  , necRevEntry(PVR_MUTE, pvr_muteCmd_B)
  , necRevEntry(PVR_RECALL, pvr_goBackCmd_B)
  , necRevEntry(PVR_INFO, pvr_infoCmd_B)
  , necRevEntry(PVR_LIST, pvr_listCmd_B)
  , necRevEntry(PVR_1, pvr_1Cmd_B)
  , necRevEntry(PVR_2, pvr_2Cmd_B)
  , necRevEntry(PVR_3, pvr_3Cmd_B)
  , necRevEntry(PVR_4, pvr_4Cmd_B)
  , necRevEntry(PVR_5, pvr_5Cmd_B)
  , necRevEntry(PVR_6, pvr_6Cmd_B)
  , necRevEntry(PVR_7, pvr_7Cmd_B)
  , necRevEntry(PVR_8, pvr_8Cmd_B)
  , necRevEntry(PVR_9, pvr_9Cmd_B)
  , necRevEntry(PVR_0, pvr_0Cmd_B)
  , necRevRepeatEntry(PVR_LEFT, pvr_leftCmd_B, 550)
  , necRevRepeatEntry(PVR_RIGHT, pvr_rightCmd_B, 550)
  , necRevRepeatEntry(PVR_UP, pvr_upCmd_B, 550)
  , necRevRepeatEntry(PVR_DOWN, pvr_downCmd_B, 550)
  , necRevEntry(PVR_OK, pvr_okCmd_B)
  , necRevEntry(PVR_STANDBY, pvr_powerCmd_B)
  , necRevEntry(PVR_TV_OUT, pvr_tvoutCmd_B)
  , necRevEntry(PVR_SETUP, pvr_setupCmd_B)
  , necRevEntry(PVR_LIVE, pvr_liveCmd_B)
  , necRevRepeatEntry(PVR_CHup, pvr_ChUpCmd_B, 550)
  , necRevRepeatEntry(PVR_CHdw, pvr_ChDnCmd_B, 550)
  , necRevEntry(PVR_FAV, pvr_favCmd_B)
  , necRevEntry(PVR_ZOOM, pvr_zoomCmd_B)
  , necRevEntry(PVR_EXIT, pvr_exitCmd_B)
  , necRevEntry(PVR_GUIDE, pvr_guideCmd_B)
};


bool haveNL = true; // set to true when have seen new line next char is Dwg char, i.e. A B etc and the following char is cmd
bool haveDwg = false; // true when have dwg char next one is cmd

char dwgChar = '\0';
char cmdChar = '\0';
uint16_t repeatInterval = 0; // set to repeat Interval for last cmd sent
IR_DATA* lastCodeSentPtr = NULL; // if NULL repeat has stopped need another code before can repeat again

unsigned long modeTimer = 0;
unsigned long MODE_TIMEOUT = 30000; // 30 sec.
bool canSetMode = true; // set to false if mode set or timeout

void setup() {
  Serial1.begin(115200);
  Serial.begin(115200);

  delay(2000); // to skip the ESP8266 debug output
  // clear input
  while (Serial1.available()) {
    Serial1.read(); // ignore Serial1 input from ESP8266
  }
  while (Serial.available()) {
    Serial.read(); // clear Serial as well
  }
  haveNL = true;
  // wait a little to let the user open the Serial output
  Serial.println();
  for (int i = 10; i >= 0; i--) {
    delay(1000);
    Serial.print(i); Serial.print(' ');
  }
  Serial.println();
  // print msg
  showMenuMsg();
  modeTimer = millis(); // start timer
  // clear input
  while (Serial1.available()) {
    Serial1.read(); // ignore Serial1 input from ESP8266
  }
  while (Serial.available()) {
    Serial.read(); // clear Serial as well
  }

}

void loop() {
  if (canSetMode) {
    if (Serial.available()) {
      char c = Serial.read();
      Serial.println(c);
      canSetMode = false;
      if ((c < '0') || (c > '2')) {
        Serial.println("Invalid mode: Only 0, 1, or 2 are valid.  Setting mode to 0");
      } else {
        mode = (c - '0');
        Serial.print("Mode set to "); Serial.println(mode);
        if (mode == 0) {
          Serial.println(" Normal operation. Take cmds from ESP8266 Wifi board (via Serial1) and send requested IR code");
        } else if (mode == 1) {
          cmdInput = &Serial;
          Serial.println(" USB commands. Take cmds from USB input (Arduino IDE Monitor at 115200) and send requested IR code.  > start repeat the last code, . stops the repeat");
        } else if (mode == 2) {
          Serial.println(" Pass Through.   USB serial connected to ESP8266 Serial for programming and debugging WiFi command output");
        } else {
          Serial.print(" Invalid mode : "); Serial.println(mode);  // should not get here!!
        }
      }
    }
    if ((millis() - modeTimer) > MODE_TIMEOUT) {
      Serial.println();
      Serial.println(" Mode setting timed out. Setting mode to 0");
      Serial.println(" Normal operation. Take cmds from ESP8266 Wifi board (via Serial1) and send requested IR code.");
      canSetMode = false;
    }
  }

  if ((mode == 0) || (mode == 1)) {
    // read serial and send IR for cmd
    while (cmdInput->available()) {
      char c = cmdInput->read();
      if ((c == '\n') || (c == '\r')) {
        haveNL = true;
        continue;
      } else if (c <= 32) {
        // skip ctrl chars and spaces
        continue;
      }
      // else not \n \r or space or ctrl
      if (!haveNL) {
        continue;  // scan for end of line
      }
      // else haveNL == true
      if (!haveDwg) {
        dwgChar = c;
        if ((dwgChar == '>') || (dwgChar == '.')) {
          // repeat last or stop repeat
          cmdChar = ' '; // dummy
          haveNL = false; // scan for end of line next time
          haveDwg = false;
        } else {
          haveDwg = true;
        }
        continue;
      }
      // else haveDwg so this char is the cmdChar
      cmdChar = c;
      haveNL = false; // scan for end of line next time
      haveDwg = false;
      break; // process this cmd
    }

    if ((dwgChar == '\0') || (cmdChar == '\0' )) {
      // nothing to do
    } else {
      // else process cmd
      if (dwgChar == '.') {
        if ((!repeatTimerRunning) || (lastCodeSentPtr == NULL))  {
          Serial.println("Received stop repeat cmd . but not repeating, so ignoring .");
        } else {
          Serial.println("Stop Repeat");
        }
        // ignore this extra repeat
        repeatTimerRunning = false;

      } else if (dwgChar == '>') {
        if (repeatTimerRunning) {
          // ignore this extra repeat
          Serial.println("Received repeat cmd > but already repeating, so ignoring > ");
        } else {
          // start repeat at repeat interval specified
          if (lastCodeSentPtr == NULL) {
            // ignore this
            Serial.println("Received repeat cmd > but no last code set, so ignoring > ");
          } else {
            // start timer for repeat
            if (lastCodeSentPtr->repeatInterval == 0) {
              Serial.println("Last cmd does not repeat, so ignoring > .");
              repeatTimerRunning = false;
            } else {
              repeatInterval = lastCodeSentPtr->repeatInterval;
              repeatTimer = millis(); // start timer
              repeatTimerRunning = true;
              Serial.print("Repeat cmd at "); Serial.print(repeatInterval); Serial.println("mS");
            }
          }
        }
      } else if (dwgChar == 'A') {
        bool found = false;
        for (size_t i = 0; i < sizeof(codes_A) / sizeof(IR_DATA); i++) {
          if (codes_A[i].cmd == cmdChar) {
            Serial.print("Block "); Serial.print(dwgChar); Serial.print(" found cmd:"); Serial.print(cmdChar); Serial.print("  Sending code : "); Serial.println(codes_A[i].codeName);
            found = true;
            lastCodeSentPtr = &(codes_A[i]);
            sendIRcode(lastCodeSentPtr->codeType, lastCodeSentPtr->data, lastCodeSentPtr->len);
          }
        }
        if (!found) {
          Serial.print("Block "); Serial.print(dwgChar); Serial.print(" cmd:"); Serial.print(cmdChar); Serial.println(" not found.");
        }

      } else if (dwgChar == 'B') {
        bool found = false;
        for (size_t i = 0; i < sizeof(codes_B) / sizeof(IR_DATA); i++) {
          if (codes_B[i].cmd == cmdChar) {
            Serial.print("Block "); Serial.print(dwgChar); Serial.print(" found cmd:"); Serial.print(cmdChar); Serial.print("  Sending code : "); Serial.println(codes_B[i].codeName);
            found = true;
            lastCodeSentPtr = &(codes_B[i]);
            sendIRcode(lastCodeSentPtr->codeType, lastCodeSentPtr->data, lastCodeSentPtr->len);
          }
        }
        if (!found) {
          Serial.print("Block "); Serial.print(dwgChar); Serial.print(" cmd:"); Serial.print(cmdChar); Serial.println(" not found.");
        }

      } else {
        // invalid do nothing
        Serial.print("Invalid Block "); Serial.print(dwgChar); Serial.print(" (cmd:"); Serial.print(cmdChar); Serial.println(")");
      }
      // clear for next time
      dwgChar = '\0';
      cmdChar = '\0';
    }

    //repeatTimerRunning = true;
    // Handle repeats, if any
    if ((repeatTimerRunning) && (millis() - repeatTimer) > repeatInterval) {
      // send repeat code
      if ((lastCodeSentPtr != NULL) && (lastCodeSentPtr->repeatInterval != 0)) {
        sendIRcode(lastCodeSentPtr->codeType, lastCodeSentPtr->data, lastCodeSentPtr->len);
        repeatTimer = millis(); // note repeat will drift by the time it takes to send the IR code
      } else {
        repeatTimerRunning = false; // no data
      }
    }
  } else if (mode == 2) {
    // mode == 2 pass through
    // for programming ESP8266 or debugging the WiFi IR command output
    if (Serial.available() > 0) {
      Serial1.write(Serial.read());
    }
    while (Serial1.available() > 0) {
      Serial.write(Serial1.read());
    }

  } else {
    mode = 0; // set to default if invalid mode
  }

}


void sendIRcode(enum codeTypesEnum codeType, uint8_t* data, unsigned int len) {
  if (codeType == RAW_code) {
    // raw first value is freq
    unsigned int* intData = (unsigned int*)data;
    unsigned int khz = *intData;  // first point is freq
    if (mode == 1) {
      Serial.print("RAW_code  Freq : "); Serial.print(khz); Serial.println("kHz");
    }
    unsigned int* irSignal = intData + 1; // skip first data point to get to pulse data
    size_t len = (len / sizeof(unsigned int)) - 1; // subtract 1 from number of unsigned ints since first point is freq
    // note code.len is always in bytes
    sendRaw(irSignal, len, khz);

  } else if (codeType == LIR_code) {
    char *codeStr = (char*)data; // null terminated
    if (mode == 1) {
      Serial.print("LIR_code  : "); Serial.println(codeStr);
    }
    int result = parseAndSendLIRcode(codeStr);
    if (result != 0) {
      if (mode == 1) {
        Serial.print("parseLIRcode returned error : "); Serial.println(result);
      }
    }

  } else if (codeType == NEC_38_code) {
    // NEC 38kHz
    uint32_t necCode = *((uint32_t*)data);
    if (mode == 1) {
      Serial.print("NEC_38 : "); Serial.println(necCode, HEX);
    }
    sendNEC(necCode, 32);

  } else if (codeType == NEC_38_REV8_code) {
    // NEC 38kHz but apply reverse8_uint32 first  BEYONDWIZ pvr needs this to convert handbook codes for IRremote sending.
    uint32_t necCode = reverse8_uint32(*((uint32_t*)data));
    if (mode == 1) {
      Serial.print("NEC_38_REV8 : "); Serial.println(*((uint32_t*)data), HEX); Serial.print(" reversed8 to send: "); Serial.println(necCode, HEX);
    }
    sendNEC(necCode, 32);

  } else if (codeType == PANASONIC38_4004_code) {
    // PANASONIC address 4004 38kHz
    uint32_t panasonicData = *((uint32_t*)data);
    if (mode == 1) {
      Serial.print("PANASONIC38_4004_code : 4004"); Serial.println(panasonicData, HEX);
    }
    sendPanasonic38(0x4004, panasonicData);

  } else if (codeType == PANASONIC38_4004_POWER_MACRO_code) { // TV_POWER and TV_MUTE is hard code here
    // PANASONIC address 4004 38kHz
    uint32_t panasonicData = *((uint32_t*)data);
    if (mode == 1) {
      Serial.print("PANASONIC38_4004_POWER_code : 4004"); Serial.println(TV_POWER, HEX);
    }
    // send power code 3 times with 74mS delay between each one
    sendPanasonic38(0x4004, TV_POWER);
    delay(74);
    sendPanasonic38(0x4004, TV_POWER);
    delay(74);
    sendPanasonic38(0x4004, TV_POWER);
    delay(5000); // wait 5sec and then mute
    sendPanasonic38(0x4004, TV_MUTE); // send mute after power (on or off)
  }
}

void showMenuMsg() {
  Serial.println(" There are 4 modes available, 0, 1, 2 and 3.  Enter a number to choose that mode.  Default is 0 if nothing entered.");
  Serial.println(" Mode selection is disabled after 30 secs.");
  Serial.println();
  Serial.println(" Mode 0 -- (default) Take cmds from ESP8266 Wifi board (via Serial1) and send requested IR code.");
  Serial.println(" Mode 1 -- Take cmds from USB serial input (Arduino IDE monitor at 115200) and send requested IR code.");
  Serial.println(" Mode 2 -- Pass through USB serial to and ESP8266 WiFi board.  Use this to program the ESP8266 and to debug the commands the WiFi sends to the Teensy.");
  Serial.print(" Enter Mode: ");
}

